import datetime

import pytest
from rest_framework.test import APIClient

from django.contrib.auth.models import Permission, User
from program.tests.factories import (
    EpisodeFactory,
    ShowFactory,
    TimeslotFactory,
    UserWithPermissionsFactory,
)

pytestmark = pytest.mark.django_db


@pytest.fixture
def _owned_show(user_with_episode_perms):
    show = ShowFactory.create()
    show.owners.set([user_with_episode_perms])
    return show


def url(episode=None):
    if episode:
        return f"/api/v1/episodes/{episode.id}/"
    else:
        return "/api/v1/episodes/"


@pytest.fixture
def user_with_episode_perms(request) -> User:
    """User with add_episode, change_episode, delete_episode permissions"""
    excluded_perms = set(getattr(request, "param", []))
    permissions = Permission.objects.filter(
        codename__in={
            "add_episode",
            "change_episode",
            "delete_episode",
            "edit__episode__title",
            "edit__episode__summary",
            "edit__episode__content",
            "edit__episode__image",
            "edit__episode__contributors",
            "edit__episode__languages",
            "edit__episode__media_description",
            "edit__episode__tags",
            "edit__episode__links",
        }
        - excluded_perms
    )

    return UserWithPermissionsFactory.create(user_permissions=permissions)


@pytest.fixture
def api_client_episode_perms(api_client, user_with_episode_perms) -> APIClient:
    """Authenticated API client for user with episode permissions"""

    api_client.force_authenticate(user_with_episode_perms)
    yield api_client
    api_client.force_authenticate()


def test_read_episodes_as_unauthenticated_user(api_client, show):
    instances = EpisodeFactory.create_batch(size=3, show=show)
    response = api_client.get(url())
    assert response.status_code == 200
    assert len(response.data) == len(instances)
    for episode_data in response.data:
        assert "memo" not in episode_data


def test_memo_is_contained_in_authenticated_episode_request(api_client_episode_perms, show):
    memo = "my memo"
    episode = EpisodeFactory.create(show=show, memo=memo)
    response = api_client_episode_perms.get(url(episode))
    assert response.status_code == 200
    assert response.data["memo"] == memo


def test_user_can_update_episode_of_owned_show(api_client_episode_perms, _owned_show):
    episode = EpisodeFactory.create(
        title="a simple title",
        summary="a simple summary",
        show=_owned_show,
    )
    data = {
        "title": "hello",
        "summary": "a new summary",
    }
    response = api_client_episode_perms.patch(url(episode), data=data)
    assert response.status_code == 200, response.content
    assert response.data["title"] == data["title"]
    assert response.data["summary"] == data["summary"]


def test_user_cannot_update_episode_of_unowned_show(api_client_episode_perms):
    show = ShowFactory.create()
    episode = EpisodeFactory.create(show=show)
    response = api_client_episode_perms.patch(url(episode), data={"title": "a new title"})
    expected_response = {
        "message": "You do not have permission to perform this action.",
        "code": "permission_denied",
    }
    assert response.status_code == 403
    assert response.json() == expected_response


def test_can_update_episode_links(api_client_episode_perms, _owned_show, link_type):
    episode = EpisodeFactory.create(show=_owned_show)
    assert episode.links.count() == 0
    data = {"links": [{"type_id": link_type.id, "url": "https://aura.radio"}]}
    response = api_client_episode_perms.patch(url(episode), data, format="json")
    assert response.status_code == 200
    assert response.data["links"] == data["links"]


@pytest.mark.parametrize("user_with_episode_perms", [["edit__episode__languages"]], indirect=True)
def test_cannot_write_episode_field_without_permission(
    api_client_episode_perms, _owned_show, language
):
    episode = EpisodeFactory.create(show=_owned_show)
    response = api_client_episode_perms.patch(
        url(episode), data={"languageIds": [language.id]}, format="json"
    )
    expected_response = {
        "languageIds": {
            "message": "You do not have permission to set this field.",
            "code": "missing-field-permission",
        }
    }
    assert response.status_code == 403
    assert response.json() == expected_response


def test_can_write_episode_field_with_permission(
    user_with_episode_perms, api_client_episode_perms, _owned_show
):
    episode = EpisodeFactory.create(show=_owned_show)
    data = {"title": "my new title"}
    response = api_client_episode_perms.patch(url(episode), data=data)
    assert response.status_code == 200
    assert response.data["title"] == data["title"]


def test_episode_updated_by_sets_updated_by_field(
    user_with_episode_perms, api_client_episode_perms, _owned_show
):
    episode = EpisodeFactory.create(show=_owned_show)
    assert episode.updated_by == ""
    assert user_with_episode_perms.username != ""
    response = api_client_episode_perms.patch(url(episode), data={})
    assert response.status_code == 200
    episode.refresh_from_db()
    assert episode.updated_by == user_with_episode_perms.username


def test_user_can_create_new_episode(api_client_episode_perms, _owned_show, profile):
    create_data = {
        "showId": _owned_show.id,
        "title": "a simple title",
        "contributorIds": [profile.id],
        "mediaDescription": "a simple media description",
    }
    response = api_client_episode_perms.post(url(), create_data, format="json")
    data = response.json()
    assert response.status_code == 201, response.content.decode()
    assert data["title"] == create_data["title"]
    assert data["mediaDescription"] == create_data["mediaDescription"]
    assert data["showId"] == _owned_show.id


def test_user_can_delete_existing_episode_of_owned_show(api_client_episode_perms, _owned_show):
    episode = EpisodeFactory.create(show=_owned_show)
    response = api_client_episode_perms.delete(url(episode))
    assert response.status_code == 204


def test_user_cannot_delete_existing_episode_of_unowned_show(api_client_episode_perms, show):
    episode = EpisodeFactory.create(show=show)
    response = api_client_episode_perms.delete(url(episode))
    assert response.status_code == 403


def test_common_user_cannot_delete_existing_show(common_api_client1, show):
    episode = EpisodeFactory.create(show=show)
    response = common_api_client1.delete(url(episode))
    assert response.status_code == 403


def test_user_cannot_move_episode_to_unowned_show(api_client_episode_perms, _owned_show, show):
    episode = EpisodeFactory.create(show=_owned_show)
    response = api_client_episode_perms.patch(url(episode), data={"showId": show.id})
    assert response.status_code == 400


def test_user_cannot_move_episode_from_unowned_show_to_owned_show(
    api_client_episode_perms, _owned_show, show
):
    episode = EpisodeFactory.create(show=show)
    response = api_client_episode_perms.patch(url(episode), data={"showId": _owned_show.id})
    assert response.status_code == 403


def test_user_with_update_episode_permission_can_update_any_episode(show, api_client):
    episode = EpisodeFactory.create(show=show)
    permissions = Permission.objects.filter(codename__in={"update_episode", "change_episode"})
    user = UserWithPermissionsFactory(user_permissions=permissions)
    api_client.force_authenticate(user)
    response = api_client.patch(url(episode), {"title": "a privileged title"})
    assert response.status_code == 200


def test_episode_timeslot_ids_are_ordered_by_start_time(
    admin_api_client, owned_show_once_timeslot
):
    schedule = owned_show_once_timeslot.schedule
    episode = EpisodeFactory(show=schedule.show)
    second_ts = owned_show_once_timeslot
    second_ts.episode = episode
    second_ts.save()

    first_ts_start = second_ts.start - datetime.timedelta(weeks=1)
    first_ts = TimeslotFactory.create(
        schedule=schedule,
        episode=episode,
        start=first_ts_start,
        end=first_ts_start + datetime.timedelta(hours=1),
    )
    third_ts_start = second_ts.start + datetime.timedelta(weeks=1)
    third_ts = TimeslotFactory.create(
        schedule=schedule,
        episode=episode,
        start=third_ts_start,
        end=third_ts_start + datetime.timedelta(hours=1),
    )

    response = admin_api_client.get(url(episode))
    # timeslot ids must be in order of start time
    assert response.json()["timeslotIds"] == [first_ts.id, second_ts.id, third_ts.id]


def test_episode_set_media_description(api_client_episode_perms, _owned_show):
    episode = EpisodeFactory.create(show=_owned_show)
    update = {"mediaDescription": "a media description"}

    response = api_client_episode_perms.patch(url(episode=episode), data=update)

    assert response.json()["mediaDescription"] == update["mediaDescription"]


def test_episode_clear_media_description(api_client_episode_perms, _owned_show):
    episode = EpisodeFactory.create(show=_owned_show)

    response = api_client_episode_perms.patch(url(episode=episode), data={"mediaDescription": ""})

    assert response.json()["mediaDescription"] == ""
