#
# steering, Programme/schedule management for AURA
#
# Copyright (C) 2011-2017, 2020, Ernesto Rico Schmidt
# Copyright (C) 2017-2019, Ingo Leindecker
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import functools
import json
from collections.abc import Iterable, MutableMapping
from datetime import date, datetime, time
from typing import Any, Callable, Dict, Hashable, Optional, Tuple, Union

import requests
from drf_spectacular.utils import OpenApiExample, extend_schema
from rest_framework import serializers
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response

from django import db
from django.conf import settings
from django.utils import timezone
from program.models import Episode, EpisodeLink, Profile, ProfileLink, Show, ShowLink
from program.typing import Link


def parse_datetime(date_string: str | None) -> datetime | None:
    """
    parse a datetime string and return a timezone aware datetime object.

    Returns `None` if no datetime string is given.
    """

    if date_string is None:
        return None

    try:
        parsed_datetime = datetime.strptime(date_string, "%Y-%m-%d %H:%M:%S")
    except ValueError:
        parsed_datetime = datetime.strptime(date_string, "%Y-%m-%d %H:%M:%S%z")

    return (
        timezone.make_aware(parsed_datetime)
        if timezone.is_naive(parsed_datetime)
        else parsed_datetime
    )


def parse_date(date_string: str | None) -> date | None:
    """
    parse a date string and return a date object.

    Returns `None` if no date string is given.
    """

    if date_string is None:
        return None

    return datetime.strptime(date_string, "%Y-%m-%d").date()


def parse_time(date_string: str | None) -> time | None:
    """
    parse a time string and return a time object.

    Returns `None` if no time string is given.
    """

    if date_string is None:
        return None

    if len(date_string) == 5:
        return datetime.strptime(date_string, "%H:%M").time()
    else:
        return datetime.strptime(date_string, "%H:%M:%S").time()


def get_audio_url(cba_id: Optional[int]) -> str:
    """
    Retrieve the direct URL to the mp3 in CBA
    In order to retrieve the URL, stations need
       - to be whitelisted by CBA
       - an API Key

    For these contact cba@fro.at
    """

    if not cba_id or settings.CBA_API_KEY == "":
        return ""
    else:
        if settings.DEBUG:
            url = (
                "https://cba.fro.at/wp-content/plugins/cba/ajax/cba-get-filename.php?post_id="
                + str(cba_id)
                + "&c=Ml3fASkfwR8"
            )
        else:
            url = (
                settings.CBA_AJAX_URL
                + "?action=cba_ajax_get_filename&post_id="
                + str(cba_id)
                + "&api_key="
                + settings.CBA_API_KEY
            )

        try:
            return requests.get(url).json()
        except (requests.RequestException, json.JSONDecodeError):
            # TODO: we might want to add some logging
            return ""


def get_values(
    kwargs: Dict[str, str], *keys: str
) -> Union[Tuple[Union[int, str, None], ...], int, str, None]:
    """Get the values of the keys from the kwargs."""

    def int_if_digit(value: Optional[str]) -> Optional[Union[int, str]]:
        return int(value) if value and value.isdigit() else value

    values = [kwargs.get(key) for key in keys]

    if len(values) > 1:
        return tuple(int_if_digit(value) for value in values)
    else:
        return int_if_digit(values[0])


def pick_keys(
    mapping: MutableMapping,
    keys: Iterable[Hashable],
    *,
    ignore_missing: bool = True,
    pop: bool = True,
):
    result = {}
    for key in keys:
        try:
            if pop:
                result[key] = mapping.pop(key)
            else:
                result[key] = mapping[key]
        except KeyError:
            if ignore_missing:
                continue
            raise
    return result


def update_links(
    instance: Union["Profile", "Episode", "Show"], links: list[Link]
) -> Union["Profile", "Episode", "Show"]:
    """Update the links associated with the instance"""

    # delete the links associated with the instance
    if instance.links.count() > 0:
        for link in instance.links.all():
            link.delete(keep_parents=True)

    if isinstance(instance, Profile):
        for link_data in links:
            ProfileLink.objects.create(profile=instance, **link_data)

    if isinstance(instance, Episode):
        for link_data in links:
            EpisodeLink.objects.create(episode=instance, **link_data)

    if isinstance(instance, Show):
        for link_data in links:
            ShowLink.objects.create(show=instance, **link_data)

    return instance


def saves_relationships_to(*field_names):
    """
    Marks a function as a relationship-updater for a serializer.

    Relationships for objects can usually only be updated once the instance has been
    saved to the database. This causes a common pattern in which
    1. the serializer relationship fields are removed from the validated_data,
    2. the instance is saved with all remaining non-relationship fields,
    3. the relationships are updated.

    We can simplify this process by marking a relationship update function
    and providing the field names it is going to handle.

    The update and create calls can then be decorated with the trigger.
    This way update and create are called once all relationship fields
    have been removed from validated data, and the relationship-updater function
    is called once update and create have finished and returned the instance.

    Example:

    .. highlight:: python
    .. code-block:: python
        class StoreSerializer(serializers.ModelSerializer)
            @saves_relationships_to("books")
            def _update_relationships(instance, books=None):
                if books:
                    instance.books.set(books)

            @_update_relationships.trigger
            def update(instance, validated_data):
                return super().update(instance, validated_data)

            @_update_relationships.trigger
            def create(validated_data):
                return super().create(validated_data)
    """
    relationship_updater: Callable[[Any, dict], None] | None = None

    def trigger_decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args):
            if len(args) == 2:
                instance, validated_data = args
            else:
                instance, validated_data = None, args[0]
            relationships = pick_keys(validated_data, field_names)
            if instance:
                saved_instance = func(self, instance, validated_data)
            else:
                saved_instance = func(self, validated_data)
            if relationships and relationship_updater:
                relationship_updater(self, saved_instance, **relationships)
            return saved_instance

        return wrapper

    def relationship_update_decorator(func):
        nonlocal relationship_updater
        relationship_updater = func

        @functools.wraps(func)
        def wrapper(self, instance, relationships):
            func(self, instance, relationships)

        setattr(wrapper, "trigger", trigger_decorator)
        return wrapper

    return relationship_update_decorator


def reorder_mixin_factory(
    order_field_name: str | None = None,
    id_field_type: type[serializers.Field] = serializers.IntegerField,
    serializer_class: type[serializers.Serializer] | None = None,
):
    """
    Factory function that creates a mixin for handling bulk reordering of objects in a ViewSet.

    :param order_field_name: the field that stores the ordinal
    :param id_field_type: the field type that stores the id
    :param serializer_class: Optional custom response serializer for the endpoint
    :returns: the parameterized mixin class for the ViewSet
    """

    class ReorderItemSerializer(serializers.Serializer):
        id = id_field_type()
        order = serializers.IntegerField(source=order_field_name)

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # Get the model from the parent's context if available
            if self.parent and getattr(self.parent, "context", None):
                viewset = self.parent.context.get("view")
                if viewset:
                    self.fields["id"].queryset = viewset.get_queryset()

    class ReorderRequestSerializer(serializers.Serializer):
        orderings = ReorderItemSerializer(many=True)

    class ReorderMixin:
        @extend_schema(
            methods=["patch"],
            request=ReorderRequestSerializer,
            responses={200: serializer_class} if serializer_class else {204: None},
            examples=[
                OpenApiExample(
                    "Reorder Example",
                    value={
                        "orderings": [
                            {"id": 1, "order": 5},
                            {"id": 2, "order": 2},
                            {"id": 3, "order": 1},
                        ]
                    },
                    description="Example request to reorder multiple objects",
                )
            ],
            description="Reorder multiple objects in a single operation.",
            summary="Bulk reorder objects",
        )
        @action(
            detail=False,
            methods=["patch"],
            url_path="reorder",
            serializer_class=ReorderRequestSerializer,
        )
        def reorder(self, request: Request) -> Response:
            serializer = ReorderRequestSerializer(data=request.data)
            serializer.is_valid(raise_exception=True)

            new_orderings = serializer.validated_data["orderings"]
            requested_obj_ids = [ordering["id"] for ordering in new_orderings]
            model = self.get_queryset().model
            obj_ids = model.objects.filter(id__in=requested_obj_ids).values_list("id", flat=True)
            update_map = {
                ordering["id"]: ordering["order"]
                for ordering in new_orderings
                if ordering["id"] in obj_ids
            }

            with db.transaction.atomic():
                objects = model.objects.select_for_update().filter(id__in=update_map.keys())

                for pk, order in update_map.items():
                    objects.filter(id=pk).update(**{order_field_name or "order": order})

                if serializer_class:
                    updated_objects = self.get_queryset().filter(id__in=update_map.keys())
                    response_serializer = serializer_class(updated_objects, many=True)
                    return Response(response_serializer.data)
                return Response(status=204)

    return ReorderMixin
